## CMSC424 Fall 2024 Assignment 1a: SQL
*The assignment is to be done by yourself, do not turn in AI-generated work.*
You **will** need to create SQL queries on the first exam.

The following assumes you have gone through PostgreSQL instructions and have ran some queries on the `university` database.

Download the Assignment 1a distribution <a href="https://ceres.cs.umd.edu/424/assign/assignment1aDist.tgz">here</a>.
The resulting files are:

1. populate.sql: The SQL script for creating the data.
1. queries.py: The file where to enter your answer
1. SQLTesting.py: File to be used for running the queries (in `queries.py`) against the database, and generate the file to be submitted.
1. Dockerfile: A modified Dockerfile for this project.

### Getting started
Start the container as in Assignment 0. Once inside the container you
would normally:
- Create a new database called `elections`
- Run `\i populate.sql` from within the `psql` interface to create and populate the tables. 
However, we have done these two steps for you in the Dockerfile
because of the ephemeral nature of the container file system (see Note below).

You never *have* to use Docker. However, it might
make things easier for you and would definitely make it easier for a
TA trying to help you.

### Schema 
The dataset contains results of `senate` and `presidential` elections for a subset of the years. For the `senate`, it contains only the statewide results from 1976 to 2018, whereas for the `presidential` elections, it contains county-level data going back to 2000.
The data was collected from https://electionlab.mit.edu/data.
The schema of the tables should be self-explanatory. 


Some things to remember: 
- The `special senate` elections are problematic. Typically senate elections take place every 6 years, with the two elections for a given state staggered. So generally speaking, any given year (say 2018), there would only be one senate election per state. However, because of special circumstances, there are
sometimes 2 elections in a given year for the same state. These two can be disambiguated based on the `specialelections` boolean flag in the database.

In many cases (especially for complex queries or queries involving 
`max` or `min`), you will find it easier to create temporary tables
using the `with` construct. This also allows you to break down the full
query and makes it easier to debug.

You don't have to use the "hints" if you don't want to; there might 
be simpler ways to solve the questions.

### Testing using SQLTesting.py
Build and run a container as before (`docker build --rm -t 424 .`):
- bash: `docker run -it -v $(pwd):/424 424`
- tcsh: ``docker run -it -v `pwd`:/424 424``.
- PowerShell: `docker run -it -v "${PWD}:/424" 424`. You might have to install WSL 2, and for
 this powershell will have to be running as administrator as well.
- windows CMD: `docker run -it -v "%cd%:/424" 424`. The above caveats might apply as well.
and you should have your files in `/424`.

You will be writing a series of queries to implement the prompts in
`queries.py`. Your answers (i.e., SQL queries) should be added to
`queries.py` file at the appropriate line. A simple query is provided
for the first answer to show you how it works.  You are also provided
with a Python file `SQLTesting.py` for testing your answers.

We recommend that you use `psql elections` to design your queries,
then paste the queries to the `queries.py` file and confirm
correctness. Use `python3 SQLTesting.py -h` to see command-line
options. For example:
- `python3 SQLTesting.py -i` -- will run each of the queries and show you your answer
- `python3 SQLTesting.py -q 1` will run question 1. 
- `python3 SQLTesting.py -i` will run all the queries, one at a time (waiting for you to press Enter after each query).


### Notes/Errata
- We grade your queries by running a modified version of `SQLTesting.py` that compares the returned answers against correct answers. So it imperative that `python3 SQLTesting.py` runs without errors.
- Database changes **do not** persist across container restarts,
i.e. across distinct run commands. This is because database state is
stored in the container's file system, which is ephemeral. This should
not be an issue as `queries.py` is mounted from the distribution directory.
- On Apple silicon you can use postgres directly via **Homebrew**:
```
   brew install postgresql@14
   brew services restart postgresql@14
```
Before running `SQLTesting.py`, you also need to install `psycopg2`:
```
   pip3 install psycopg2
```

## Submission Instructions
See `queries.py` for queries to write.
Submit your altered `queries.py` to Gradescope at <a href="https://www.gradescope.com/courses/811728/assignments/4669989">Assignment 1a</a>.
      
## Relevance

Not only will this inform our later discussion of implementation
strategies, but the first test will either require you to write SQL
queries along these lines (w/ the elections databases), or critique
given queries.
