## Assignment 6: Snapshot Isolation

*The assignment is to be done by yourself.*

### Setup

Download files for Assignment 6 [here](https://ceres.cs.umd.edu/424/assign/assignment6Dist.tgz?1").
Build (`docker build --rm -t 424 .`) and run (`docker run -it -v
$(pwd):/424 424`) to create your container.



### Overview

In this assignment you will modify our simple database to emulate first-committer *snapshot
isolation* instead of lock-based two-phase consistency.

Snapshot isolation is usually implemented using a *multi-versioned* database;
snapshot reads can be done at any time by specifying older timestamps. We do
not have a multi-versioned database, so instead we will handle snapshot reads
by specifying the readset and writeset ahead of time, and *stashing* a snapshot of the
relevant data. Note that this technique is also used in some high-performance
distributed databases<sup>1</sup>. Subsequent reads by the transaction will read from the stashed
copies, rather than directly from the database.

Writes in snapshot isolation are buffered, and only pushed to the database
when the transaction *validates*, i.e. is guaranteed to succeed. A transaction *i*
validates if and only if no other transaction commits between the time *i*
starts and attempts to validate. This is straightforward in a multi-versioned
database which, again, we do not have. We will emulate this by allowing *i* to validate if none of
the items in its writeset change values during the course of *i*'s execution.
This essentially means that (1) we also stash all data in the writeset at the
start of transaction execution, and (2) we compare the stashed value w/ the
"current" value at commit time to see if it has changed.

Transactions that fail validation are aborted.

Note that in a real database validation and commitment should be a single
atomic action; we will not go that far in this system. The example
transactions should never execute their transactions close enough together for
this discrepancy to be apparent.

### Details.
As in prior assignments, you will only have to write a small amount of
code. Your changes will be confined to `transactions.py`, and *only* that file
should be uploaded to gradescope.  Search for "SNAPSHOT_ISO" or "YOUR CODE
HERE" to find places that need to be modified.

The following is an example transaction:
```
    tstate = TransactionState()
    tstate.setMode(TransactionState.SNAPSHOT_ISO)
    tstate.takeSnapshot(relation, "A", {primary_id1, primary_id2})
    v = tstate.getAttribute(relation, primary_id2, attr)
    writeVal = str(100 - int(v))
    tstate.setAttribute(relation, primary_id1, attr, writeVal)
    res = tstate.commitTransaction()
```

Our approach will be the following:
1. The transaction explicitly takes a snapshot of all values in
the readset and writeset. Real systems use a multi-versioned database to access the
correct data. 
1. Reads are satisfied from the snapshot, *or writes from the same transaction*.
2. Written values are tracked and added to a writebuffer
rather than being written to the relation.
3. At commit time, we check to see whether data for which we have updates in our writebuffers
has been changed in the relation by any other transaction. Since our system does not currently have
explicit versioning,    
we approximate this by checking to see whether the any *values* of the data we are writing
have changed.

Logically, you will have to add three pieces of state to the
transaction's initialization:
- *snapshot*: a `dict` of database values indexed by `(attr, id)`
- *writeBuffer*: an ordered list of values to be written out at
successfull commit
- *writeset*: a set describing which data has been modified. This can be
extracted from the writeBuffer at commit time, and so does not have to
be tracked explicitly. 

You do not have to structure your data this way, and there is some
redundancy in the above.

### Testing
Use `python3 testing.py` to test your implementation locally prior to
gradescope submission. Remove `relation1` before testing:
```
    rm relation1 ; python3 testing.py
```

### Submission
Submit the modified `transactions.py` file to [Gradescope](https://www.gradescope.com/courses/811728/assignments/4828403/review_grades). 

You should not change any other files.

<sup>1</sup>Thomson, A., Diamond, T., Weng, S. C., Ren, K., Shao, P., & Abadi, D. J. "Calvin: fast distributed transactions for partitioned database systems." *Proceedings of the 2012 ACM SIGMOD International Conference on Management of Data*. 2012.

